/*
 *  Anticipatory Deadline i/o scheduler.
 *  Derived from Deadline i/o scheduler.
 *  Proactively checks if any I/O request is going to miss the deadline.
 *  If so serves that request immediately out of order.
 */
#include <linux/kernel.h>
#include <linux/fs.h>
#include <linux/blkdev.h>
#include <linux/elevator.h>
#include <linux/bio.h>
#include <linux/module.h>
#include <linux/slab.h>
#include <linux/init.h>
#include <linux/compiler.h>
#include <linux/hash.h>
#include <linux/rbtree.h>
#include <linux/tivo-monotime.h>
#include <linux/ide-tivo.h>

#undef DEBUG_AD_SCHED
#ifdef DEBUG_AD_SCHED
#define DPRINTK(fmt, args...)           printk(fmt, ## args);      
#else
#define DPRINTK(fmt, args...)
#endif

#define RESTORE_ELV_RQ_ORDER    1

static const int ad_hash_shift = 5;
#define DL_HASH_BLOCK(sec)	((sec) >> 3)
#define DL_HASH_FN(sec)		(hash_long(DL_HASH_BLOCK((sec)), ad_hash_shift))
#define DL_HASH_ENTRIES		(1 << ad_hash_shift)
#define rq_hash_key(rq)		((rq)->sector + (rq)->nr_sectors)
#define ON_HASH(drq)		(!hlist_unhashed(&(drq)->hash))

struct ad_data {
	/*
	 * run time data
	 */

	/*
	 * requests (ad_rq s) are present on both sort_list and fifo_list
	 */
	struct rb_root sort_list;	
	struct list_head fifo_list;
	
	/*
	 * next in sort order. 
	 */
	struct ad_rq *next_drq;
	struct hlist_head *hash;	/* request hash */
	sector_t last_sector;		/* head position */

	/*
	 * settings that change how the i/o scheduler behaves
	 */
	int front_merges;
        int seek; /* Disk Avg seek time in msec */
        int rotation;/* Disk Avg Rotation time in msec */
        int transfer; /* Worst DTR in msec */

	mempool_t *drq_pool;
};

/*
 * pre-request data.
 */
struct ad_rq {
	/*
	 * rbtree index, key is the starting offset
	 */
	struct rb_node rb_node;
	sector_t rb_key;

	struct request *request;

	/*
	 * request hash, key is the ending offset (for back merge lookup)
	 */
	struct hlist_node hash;

	/*
	 * expire fifo
	 */
	struct list_head fifo;
};

static void ad_move_request(struct ad_data *dd, struct ad_rq *drq);

static kmem_cache_t *drq_pool;

#define RQ_DATA(rq)	((struct ad_rq *) (rq)->elevator_private)

/* Deadline estimation helper functions */
static inline unsigned long ad_estimate( struct ad_data *dd,
                    /*unsigned long current_sector,*/
                    struct request *req) 
{
        unsigned long result;

        result = dd->seek + /* initial seek time */
                dd->rotation + /* initial rotation time */ 
                (dd->transfer * req->nr_sectors) + /* transfer time */
                ((req->nr_sectors>>8) * dd->rotation); /* extra rotations */

        return result;
}

static inline unsigned long ad_estimate_noseek( struct ad_data *dd,
                       /*unsigned long current_sector,*/
                       struct request *req) 
{
        unsigned long result;

        /* Same, but don't include seek time (because it will be included
         in a subsequent time estimate)*/
        result = dd->rotation + /* initial rotation time */
                (dd->transfer * req->nr_sectors) + /* transfer time */
                ((req->nr_sectors>>8) * dd->rotation); /* extra rotations */

        return result;
}

/*
 * the back merge hash support functions
 */
static inline void __ad_del_drq_hash(struct ad_rq *drq)
{
	hlist_del_init(&drq->hash);
}

static inline void ad_del_drq_hash(struct ad_rq *drq)
{
	if (ON_HASH(drq))
		__ad_del_drq_hash(drq);
}

static inline void
ad_add_drq_hash(struct ad_data *dd, struct ad_rq *drq)
{
	struct request *rq = drq->request;

	BUG_ON(ON_HASH(drq));

	hlist_add_head(&drq->hash, &dd->hash[DL_HASH_FN(rq_hash_key(rq))]);
}

/*
 * move hot entry to front of chain
 */
static inline void
ad_hot_drq_hash(struct ad_data *dd, struct ad_rq *drq)
{
	struct request *rq = drq->request;
	struct hlist_head *head = &dd->hash[DL_HASH_FN(rq_hash_key(rq))];

	if (ON_HASH(drq) && &drq->hash != head->first) {
		hlist_del(&drq->hash);
		hlist_add_head(&drq->hash, head);
	}
}

static struct request *
ad_find_drq_hash(struct ad_data *dd, sector_t offset)
{
	struct hlist_head *hash_list = &dd->hash[DL_HASH_FN(offset)];
	struct hlist_node *entry, *next;
	struct ad_rq *drq;

	hlist_for_each_entry_safe(drq, entry, next, hash_list, hash) {
		struct request *__rq = drq->request;

		BUG_ON(!ON_HASH(drq));

		if (!rq_mergeable(__rq)) {
			__ad_del_drq_hash(drq);
			continue;
		}

		if (rq_hash_key(__rq) == offset)
			return __rq;
	}

	return NULL;
}

/*
 * rb tree support functions
 */
#define rb_entry_drq(node)	rb_entry((node), struct ad_rq, rb_node)
#define DRQ_RB_ROOT(dd, drq)	(&(dd)->sort_list)
#define rq_rb_key(rq)		(rq)->sector

static struct ad_rq *
__ad_add_drq_rb(struct ad_data *dd, struct ad_rq *drq)
{
	struct rb_node **p = &DRQ_RB_ROOT(dd, drq)->rb_node;
	struct rb_node *parent = NULL;
	struct ad_rq *__drq;

	while (*p) {
		parent = *p;
		__drq = rb_entry_drq(parent);

		if (drq->rb_key < __drq->rb_key)
			p = &(*p)->rb_left;
		else if (drq->rb_key > __drq->rb_key)
			p = &(*p)->rb_right;
		else
			return __drq;
	}

	rb_link_node(&drq->rb_node, parent, p);
	return NULL;
}

static void
ad_add_drq_rb(struct ad_data *dd, struct ad_rq *drq)
{
	struct ad_rq *__alias;

	drq->rb_key = rq_rb_key(drq->request);

retry:
	__alias = __ad_add_drq_rb(dd, drq);
	if (!__alias) {
		rb_insert_color(&drq->rb_node, DRQ_RB_ROOT(dd, drq));
		return;
	}

	ad_move_request(dd, __alias);
	goto retry;
}

static inline void
ad_del_drq_rb(struct ad_data *dd, struct ad_rq *drq)
{
	if (dd->next_drq == drq) {
		struct rb_node *rbnext = rb_next(&drq->rb_node);

		dd->next_drq = NULL;
		if (rbnext)
			dd->next_drq = rb_entry_drq(rbnext);
	}

	BUG_ON(!RB_EMPTY_NODE(&drq->rb_node));
	rb_erase(&drq->rb_node, DRQ_RB_ROOT(dd, drq));
	RB_CLEAR_NODE(&drq->rb_node);
}

static struct request *
ad_find_drq_rb(struct ad_data *dd, sector_t sector)
{
	struct rb_node *n = dd->sort_list.rb_node;
	struct ad_rq *drq;

	while (n) {
		drq = rb_entry_drq(n);

		if (sector < drq->rb_key)
			n = n->rb_left;
		else if (sector > drq->rb_key)
			n = n->rb_right;
		else
			return drq->request;
	}

	return NULL;
}

/*
 * ad_find_first_drq finds the first (lowest sector numbered) request
 * Used to sweep back to the start of the disk
 * (1-way elevator) after we process the last (highest sector) request.
 */
static struct ad_rq *
ad_find_first_drq(struct ad_data *dd)
{
	struct rb_node *n = dd->sort_list.rb_node;

	for (;;) {
		if (n->rb_left == NULL)
			return rb_entry_drq(n);
		
		n = n->rb_left;
	}
}

#ifdef RESTORE_ELV_RQ_ORDER
/*
 * returns a particular ad_rq by looking at its rb_key
 * or request start sector.
 */
static struct ad_rq *
ad_find_drq(struct ad_data *dd, sector_t sector)
{
	struct rb_node *n = dd->sort_list.rb_node;
	struct ad_rq *drq;

	while (n) {
		drq = rb_entry_drq(n);

		if (sector < drq->rb_key)
			n = n->rb_left;
		else if (sector > drq->rb_key)
			n = n->rb_right;
		else
			return drq;
	}

	return NULL;
}
#endif

/*
 * add drq to rbtree and fifo
 */
static void
ad_add_request(struct request_queue *q, struct request *rq)
{
	struct ad_data *dd = q->elevator->elevator_data;
	struct ad_rq *drq = RQ_DATA(rq);
        struct list_head *list;

	ad_add_drq_rb(dd, drq);

        /* all VFS requests and few MFS requests with infinite deadline */
        if ( rq->deadline == INFINITE_DEADLINE ){
                list_add_tail(&drq->fifo,&dd->fifo_list);
        }
        else if ( !rq->deadline ){  /* MFS requests with deadline value = 0 */
                rq->deadline = INFINITE_DEADLINE;
                list_add_tail(&drq->fifo,&dd->fifo_list);
        }
        else{ /* media requests with valid deadline value */
                list_for_each(list, &dd->fifo_list) {
                        struct ad_rq *tdrq;
                        tdrq = list_entry(list, struct ad_rq, fifo);
                        if ( rq->deadline < tdrq->request->deadline)
                                break;
                }
                __list_add(&drq->fifo,list->prev,list);
        }

	if (rq_mergeable(rq))
		ad_add_drq_hash(dd, drq);
}

/*
 * remove rq from rbtree, fifo, and hash
 */
static void ad_remove_request(request_queue_t *q, struct request *rq)
{
	struct ad_rq *drq = RQ_DATA(rq);
	struct ad_data *dd = q->elevator->elevator_data;

	list_del_init(&drq->fifo);
	ad_del_drq_rb(dd, drq);
	ad_del_drq_hash(drq);
}

static int
ad_merge(request_queue_t *q, struct request **req, struct bio *bio)
{
	struct ad_data *dd = q->elevator->elevator_data;
	struct request *__rq;
	int ret;

	/*
	 * see if the merge hash can satisfy a back merge
	 */
	__rq = ad_find_drq_hash(dd, bio->bi_sector);
	if (__rq) {
		BUG_ON(__rq->sector + __rq->nr_sectors != bio->bi_sector);

		if (elv_rq_merge_ok(__rq, bio)) {
			ret = ELEVATOR_BACK_MERGE;
			goto out;
		}
	}

	/*
	 * check for front merge
	 */
	if (dd->front_merges) {
		sector_t rb_key = bio->bi_sector + bio_sectors(bio);

		__rq = ad_find_drq_rb(dd, rb_key);
		if (__rq) {
			BUG_ON(rb_key != rq_rb_key(__rq));

			if (elv_rq_merge_ok(__rq, bio)) {
				ret = ELEVATOR_FRONT_MERGE;
				goto out;
			}
		}
	}

	return ELEVATOR_NO_MERGE;
out:
	if (ret)
		ad_hot_drq_hash(dd, RQ_DATA(__rq));
	*req = __rq;
	return ret;
}

static void ad_merged_request(request_queue_t *q, struct request *req)
{
	struct ad_data *dd = q->elevator->elevator_data;
	struct ad_rq *drq = RQ_DATA(req);

	/*
	 * hash always needs to be repositioned, key is end sector
	 */
	ad_del_drq_hash(drq);
	ad_add_drq_hash(dd, drq);

	/*
	 * if the merge was a front merge, we need to reposition request
	 */
	if (rq_rb_key(req) != drq->rb_key) {
		ad_del_drq_rb(dd, drq);
		ad_add_drq_rb(dd, drq);
	}
}

static void
ad_merged_requests(request_queue_t *q, struct request *req,
			 struct request *next)
{
	struct ad_data *dd = q->elevator->elevator_data;
	struct ad_rq *drq = RQ_DATA(req);
	struct ad_rq *dnext = RQ_DATA(next);
        monotonic_t nts,ts;

	BUG_ON(!drq);
	BUG_ON(!dnext);

	/*
	 * reposition drq (this is the merged request) in hash, and in rbtree
	 * in case of a front merge
	 */
	ad_del_drq_hash(drq);
	ad_add_drq_hash(dd, drq);

	if (rq_rb_key(req) != drq->rb_key) {
		ad_del_drq_rb(dd, drq);
		ad_add_drq_rb(dd, drq);
	}

	/*
	 * if dnext expires before drq, assign its expire time to drq
	 * and move into dnext position (dnext will be deleted) in fifo
	 */
	if (!list_empty(&drq->fifo) && !list_empty(&dnext->fifo)) {
                tivo_monotonic_store(&ts, drq->request->deadline);
                tivo_monotonic_store(&nts, dnext->request->deadline);
		if (tivo_monotonic_lt(&nts,&ts)) {
			list_move(&drq->fifo, &dnext->fifo);
			drq->request->deadline = dnext->request->deadline;
		}
	}

	/*
	 * kill knowledge of next, this one is a goner
	 */
	ad_remove_request(q, next);
}

/*
 * move request from sort list to dispatch queue.
 */
static inline void
ad_move_to_dispatch(struct ad_data *dd, struct ad_rq *drq)
{
	request_queue_t *q = drq->request->q;

	ad_remove_request(q, drq->request);
	elv_dispatch_add_tail(q, drq->request);
}

/*
 * move an entry to dispatch queue
 */
static void
ad_move_request(struct ad_data *dd, struct ad_rq *drq)
{
	struct rb_node *rbnext;

#ifdef RESTORE_ELV_RQ_ORDER
        sector_t old_elv_pos = 0;

        /*
         * if the deadlined request is same as current elevator order
         * request, skip below old_elv_pos calculation etc ..
         */
        if((dd->next_drq != NULL) && (dd->next_drq != drq)) { 
                /* 
                 * save elevator rb_key(sector offset) of the next request 
                 * in elevator order before jumping to deadlined request.
                 */
                if(blk_fs_request(dd->next_drq->request))
                        old_elv_pos = dd->next_drq->rb_key;

                DPRINTK("ele rq: old_elv_pos = 0x%lx , sector = 0x%lx, nr_sectors = 0x%lx,"
                                " deadline = 0x%llx\n", old_elv_pos, dd->next_drq->request->sector,
                                dd->next_drq->request->nr_sectors, dd->next_drq->request->deadline);

                DPRINTK("deadlined rq: sector = 0x%lx, nr_sectors = 0x%lx,"
                                " deadline = 0x%llx\n", drq->request->sector, 
                                drq->request->nr_sectors, drq->request->deadline);
        }
#endif

        rbnext = rb_next(&drq->rb_node);
	dd->next_drq = NULL;

	if (rbnext)
		dd->next_drq = rb_entry_drq(rbnext);
	
	dd->last_sector = drq->request->sector + drq->request->nr_sectors;

	/*
	 * take it off the sort and fifo list, move
	 * to dispatch queue
	 */

        DPRINTK("Dispatching:: sector = 0x%lx, nr_sectors = 0x%lx, deadline = 0x%llx\n\n", 
                        drq->request->sector, drq->request->nr_sectors, drq->request->deadline);

	ad_move_to_dispatch(dd, drq);

#ifdef RESTORE_ELV_RQ_ORDER
        if(old_elv_pos){
                /*
                 * restore the elevator where it was before 
                 * jumping to out-of-order deadlined request.
                 */
                dd->next_drq = ad_find_drq(dd, old_elv_pos);
                old_elv_pos = 0;
        }
#endif        
}

#define list_entry_fifo(ptr)	list_entry((ptr), struct ad_rq, fifo)

/*
 * ad_check_fifo returns 0 if there are no expirable request on the fifo,
 * 1 otherwise. Requires !list_empty(&dd->fifo_list)
 */
static inline int ad_check_fifo(struct ad_data *dd,struct ad_rq *ndrq)
{
	struct ad_rq *drq = list_entry_fifo(dd->fifo_list.next);
        monotonic_t nts,ts;

        /*
         * avoid extra fudge in estimating the expiry condition for 
         * db requests with infinite deadline. 
         */
        if((drq->request->deadline == INFINITE_DEADLINE)){
                return 0;
        }

        tivo_monotonic_get_current(&nts);
        tivo_monotonic_add_usecs(&nts,ad_estimate(dd, drq->request));

        tivo_monotonic_store(&ts, drq->request->deadline);

	/*
	 * drq can expire
	 */
	if (tivo_monotonic_lt(&ts,&nts))
		return 1;

	return 0;
}

/*
 * ad_dispatch_requests selects the best request according to
 * read/write expire etc
 */
static int ad_dispatch_requests(request_queue_t *q, int force)
{
	struct ad_data *dd = q->elevator->elevator_data;
	const int reads = !list_empty(&dd->fifo_list);
	struct ad_rq *drq;

	if (!reads) {
	    return 0;
	}

        BUG_ON(RB_EMPTY_ROOT(&dd->sort_list));
    
	/*
	 * find best request 
	 */
	if (ad_check_fifo(dd, NULL)) {
		/* An expirable request exists - satisfy it */
		drq = list_entry_fifo(dd->fifo_list.next);
		
	} else if (dd->next_drq) {
		/*
		 * The last req was the same dir and we have a next request in
		 * sort order. No expirable requests so continue on from here.
		 */
		drq = dd->next_drq;
	} else {
		/*
		 * The last req was the other direction or we have run out of
		 * higher-sectored requests. Go back to the lowest sectored
		 * request (1 way elevator) and start a new request.
		 */
		drq = ad_find_first_drq(dd);
	}

	/*
	 * drq is the selected appropriate request.
	 */
	ad_move_request(dd, drq);

	return 1;
}

static int ad_queue_empty(request_queue_t *q)
{
	struct ad_data *dd = q->elevator->elevator_data;

	return list_empty(&dd->fifo_list);
}

static struct request *
ad_former_request(request_queue_t *q, struct request *rq)
{
	struct ad_rq *drq = RQ_DATA(rq);
	struct rb_node *rbprev = rb_prev(&drq->rb_node);

	if (rbprev)
		return rb_entry_drq(rbprev)->request;

	return NULL;
}

static struct request *
ad_latter_request(request_queue_t *q, struct request *rq)
{
	struct ad_rq *drq = RQ_DATA(rq);
	struct rb_node *rbnext = rb_next(&drq->rb_node);

	if (rbnext)
		return rb_entry_drq(rbnext)->request;

	return NULL;
}

static void ad_exit_queue(elevator_t *e)
{
	struct ad_data *dd = e->elevator_data;

	BUG_ON(!list_empty(&dd->fifo_list));

	mempool_destroy(dd->drq_pool);
	kfree(dd->hash);
	kfree(dd);
}

/*
 * initialize elevator private data (ad_data), and alloc a drq for
 * each request on the free lists
 */
static void *ad_init_queue(request_queue_t *q, elevator_t *e)
{
	struct ad_data *dd;
	int i;

	if (!drq_pool)
		return NULL;

	dd = kmalloc_node(sizeof(*dd), GFP_KERNEL, q->node);
	if (!dd)
		return NULL;
	memset(dd, 0, sizeof(*dd));

	dd->hash = kmalloc_node(sizeof(struct hlist_head)*DL_HASH_ENTRIES,
				GFP_KERNEL, q->node);
	if (!dd->hash) {
		kfree(dd);
		return NULL;
	}

	dd->drq_pool = mempool_create_node(BLKDEV_MIN_RQ, mempool_alloc_slab,
					mempool_free_slab, drq_pool, q->node);
	if (!dd->drq_pool) {
		kfree(dd->hash);
		kfree(dd);
		return NULL;
	}

	for (i = 0; i < DL_HASH_ENTRIES; i++)
		INIT_HLIST_HEAD(&dd->hash[i]);

	INIT_LIST_HEAD(&dd->fifo_list);
	dd->sort_list = RB_ROOT;
	dd->front_merges = 1;

        /* Worst case seek + rotation and data transfer times from actual data.*/
        dd->seek = 19000; /* 19ms worst case seek (no rotation) */
        dd->rotation = 11111; /* 5400rpm = 90rps = 11111us/rev */
        dd->transfer = (20000/256); /* worst case DTR -- 20ms for 256 sectors*/

	return dd;
}

static void ad_put_request(request_queue_t *q, struct request *rq)
{
	struct ad_data *dd = q->elevator->elevator_data;
	struct ad_rq *drq = RQ_DATA(rq);

	mempool_free(drq, dd->drq_pool);
	rq->elevator_private = NULL;
}

static int
ad_set_request(request_queue_t *q, struct request *rq, struct bio *bio,
		     gfp_t gfp_mask)
{
	struct ad_data *dd = q->elevator->elevator_data;
	struct ad_rq *drq;

	drq = mempool_alloc(dd->drq_pool, gfp_mask);
	if (drq) {
		memset(drq, 0, sizeof(*drq));
		RB_CLEAR_NODE(&drq->rb_node);
		drq->request = rq;

		INIT_HLIST_NODE(&drq->hash);

		INIT_LIST_HEAD(&drq->fifo);

		rq->elevator_private = drq;
		return 0;
	}

	return 1;
}

/*
 * sysfs parts below
 */

static ssize_t
ad_var_show(int var, char *page)
{
	return sprintf(page, "%d\n", var);
}

static ssize_t
ad_var_store(int *var, const char *page, size_t count)
{
	char *p = (char *) page;

	*var = simple_strtol(p, &p, 10);
	return count;
}

#define SHOW_FUNCTION(__FUNC, __VAR)				        \
static ssize_t __FUNC(elevator_t *e, char *page)			\
{									\
	struct ad_data *dd = e->elevator_data;			        \
	int __data = __VAR;						\
	return ad_var_show(__data, (page));			        \
}
SHOW_FUNCTION(ad_front_merges_show, dd->front_merges);
SHOW_FUNCTION(ad_seek_show, dd->seek);
SHOW_FUNCTION(ad_rotation_show, dd->rotation);
SHOW_FUNCTION(ad_transfer_show, dd->transfer);
#undef SHOW_FUNCTION

#define STORE_FUNCTION(__FUNC, __PTR)			                \
static ssize_t __FUNC(elevator_t *e, const char *page, size_t count)	\
{									\
	struct ad_data *dd = e->elevator_data;			        \
	int __data;							\
	int ret = ad_var_store(&__data, (page), count);		        \
        *(__PTR) = __data;					        \
	return ret;							\
}
STORE_FUNCTION(ad_front_merges_store, &dd->front_merges);
STORE_FUNCTION(ad_seek_store, &dd->seek);
STORE_FUNCTION(ad_rotation_store, &dd->rotation);
STORE_FUNCTION(ad_transfer_store, &dd->transfer);
#undef STORE_FUNCTION

#define DD_ATTR(name) \
	__ATTR(name, S_IRUGO|S_IWUSR, ad_##name##_show, \
				      ad_##name##_store)

static struct elv_fs_entry ad_attrs[] = {
	DD_ATTR(front_merges),
	DD_ATTR(seek),
	DD_ATTR(rotation),
	DD_ATTR(transfer),
	__ATTR_NULL
};

static struct elevator_type iosched_deadline = {
	.ops = {
		.elevator_merge_fn = 		ad_merge,
		.elevator_merged_fn =		ad_merged_request,
		.elevator_merge_req_fn =	ad_merged_requests,
		.elevator_dispatch_fn =		ad_dispatch_requests,
		.elevator_add_req_fn =		ad_add_request,
		.elevator_queue_empty_fn =	ad_queue_empty,
		.elevator_former_req_fn =	ad_former_request,
		.elevator_latter_req_fn =	ad_latter_request,
		.elevator_set_req_fn =		ad_set_request,
		.elevator_put_req_fn = 		ad_put_request,
		.elevator_init_fn =		ad_init_queue,
		.elevator_exit_fn =		ad_exit_queue,
	},

	.elevator_attrs = ad_attrs,
	.elevator_name = "ad",
	.elevator_owner = THIS_MODULE,
};

static int __init ad_init(void)
{
	int ret;

	drq_pool = kmem_cache_create("ad_drq", sizeof(struct ad_rq),
				     0, 0, NULL, NULL);

	if (!drq_pool)
		return -ENOMEM;

	ret = elv_register(&iosched_deadline);
	if (ret)
		kmem_cache_destroy(drq_pool);

	return ret;
}

static void __exit ad_exit(void)
{
	kmem_cache_destroy(drq_pool);
	elv_unregister(&iosched_deadline);
}

module_init(ad_init);
module_exit(ad_exit);

MODULE_LICENSE("GPL");
MODULE_DESCRIPTION("anticipatory deadline IO scheduler");
