/*
*
* SCSI Raw Disk DIO Interface
*
*/

#include <linux/bio.h>
#include <linux/genhd.h>
#include <linux/errno.h>
#include <linux/blkdev.h>
#include <linux/time.h>

#include <linux/ide-tivo.h>
#include <linux/tivo-tagio.h>
#include <linux/tivo-monotime.h>

#include <asm/uaccess.h>
#include <asm/cacheflush.h>

#include <scsi/scsi.h>
#include <scsi/scsi_cmnd.h>
#include <scsi/scsi_ioctl.h>
#include <scsi/sg.h>
#include "scsi_logging.h"

#define MAX_NRSECTORS   512
#define SECTOR_BYTES    512
#define MAX_BYTES       (MAX_NRSECTORS * SECTOR_BYTES)
unsigned int err_count = 10;

static int split_requests(struct file *filp, struct gendisk *bd_disk, const struct FsIovec *buf,
                int buf_len, unsigned long long start_sect,
                monotonic_t *deadline, int rw);

static void prepare_scsi_request(struct sg_io_hdr *sg_hdr, unsigned long long sector,unsigned long long nr_sectors,
                const struct FsIovec *iov_user, int len, int direction);


static inline long long now_usec(void) {
        long long msec;
        struct timeval tv;

        do_gettimeofday(&tv);
        msec = (long long)tv.tv_sec * 1000000 + tv.tv_usec;
        return msec; 
}

static int scsi_dosectors_io(struct file *file, struct gendisk *bd_disk, 
                struct sg_io_hdr *hdr, unsigned int sector,unsigned int nr_sectors,
                monotonic_t *deadline)
{
	unsigned long start_time, flags;
	int writing = 0, ret = 0;
	struct request *rq;
        request_queue_t *q;
	struct bio *bio;
        DECLARE_COMPLETION_ONSTACK(wait);

        q = bd_disk->queue;
	if (!q)
		return -ENXIO;

	if (hdr->cmd_len > BLK_MAX_CDB)
		return -EINVAL;

        /* converting nr_sectors to bytes */
        hdr->dxfer_len *= queue_hardsect_size(q);
        
	if (hdr->dxfer_len > (q->max_hw_sectors << 9))
		return -EIO;

	if (hdr->dxfer_len)
		switch (hdr->dxfer_direction) {
                        default:
                                return -EINVAL;
                        case SG_DXFER_TO_DEV:
                                writing = 1;
                                break;
                        case SG_DXFER_TO_FROM_DEV:
                        case SG_DXFER_FROM_DEV:
                                break;
		}

	rq = blk_get_request(q, writing ? WRITE : READ, GFP_KERNEL);
	if (!rq)
		return -ENOMEM;

        BUG_ON(!hdr->dxferp);

        ret = blk_rq_map_user_iov(q, rq, hdr->dxferp, hdr->iovec_count);

        flush_dcache_all();

	if (ret < 0){
                printk("scsi_dosectors: blk_rq_map_user_iov failed err = %d \n", ret);
		goto out;
        }

	/*
	 * fill in request structure
	 */
	rq->cmd_len = hdr->cmd_len;
	memset(rq->cmd, 0, BLK_MAX_CDB); /* ATAPI hates garbage after CDB */
	memcpy(rq->cmd, hdr->cmdp, hdr->cmd_len);

	rq->sense = hdr->sbp;
	rq->sense_len = 0;

	bio = rq->bio;

        rq->hard_sector = rq->sector = sector;
        rq->hard_nr_sectors = rq->nr_sectors = rq->current_nr_sectors = nr_sectors;

	/*
	 * bounce this after holding a reference to the original bio, it's
	 * needed for proper unmapping
	 */
	if (rq->bio)
		blk_queue_bounce(q, &rq->bio);

	rq->timeout = (hdr->timeout * HZ) / 1000;
	if (!rq->timeout)
		rq->timeout = q->sg_timeout;
#define BLK_DEFAULT_TIMEOUT  (60 * HZ)
	if (!rq->timeout)
		rq->timeout = BLK_DEFAULT_TIMEOUT;

        rq->deadline = tivo_monotonic_retrieve(deadline);

	rq->retries = 0;

	start_time = jiffies;

        /*
         * we need an extra reference to the request, so we can look at
         * it after io completion
         */
        rq->ref_count++;

        BUG_ON(!rq->sense);
        
        rq->flags |= REQ_CMD;
        rq->waiting = &wait;
        rq->rq_disk = bd_disk;
        rq->end_io = blk_end_sync_rq;
        WARN_ON(irqs_disabled());

        spin_lock_irqsave(q->queue_lock, flags);
        if (elv_queue_empty(q))
                blk_plug_device(q);
        
        __elv_add_request(q, rq, ELEVATOR_INSERT_SORT, 0);

        if (blk_queue_plugged(q))
                __generic_unplug_device(q);
        else
                q->request_fn(q);
        spin_unlock_irqrestore(q->queue_lock, flags);

        wait_for_completion(&wait);

        rq->waiting = NULL;

	hdr->status = 0xff & rq->errors;
	hdr->masked_status = status_byte(rq->errors);
	hdr->msg_status = msg_byte(rq->errors);
	hdr->host_status = host_byte(rq->errors);
	hdr->driver_status = driver_byte(rq->errors);
	hdr->info = 0;
	if (hdr->masked_status || hdr->host_status || hdr->driver_status)
		hdr->info |= SG_INFO_CHECK;
	hdr->resid = rq->data_len;
	hdr->duration = ((jiffies - start_time) * 1000) / HZ;
	hdr->sb_len_wr = 0;

        /*Return number of bytes transfered*/
        ret = hdr->resid;

	if (rq->sense_len && hdr->sbp) {
		int len = min((unsigned int) hdr->mx_sb_len, rq->sense_len);
                hdr->sb_len_wr = len;
	}

	if (blk_rq_unmap_user(bio, hdr->dxfer_len))
		ret = -EFAULT;

        flush_dcache_all();
	/* may not have succeeded, but output values written to control
	 * structure (struct sg_io_hdr).  */
out:
	blk_put_request(rq);
	return ret;
}

int scsi_dosectors(struct inode *inode,struct gendisk *bd_disk,
			     struct file *filp, void *arg)
{
        struct {
                tivotag tagtuple;
                int present;
        } tagarray[TIVOTAGIO_NUMTAGS];
        tivotag tagtuple;
        char *argPtr;
        int nTags;
        const struct FsIovec *iov_user;
        int user_len, nbytes = 0;
        int rw;
        int minor, part;
        sector_t start_sect, nr_sects;
        unsigned long long sector, nr_sectors; 
        struct sg_io_hdr sghdr;
        monotonic_t deadline;

        minor = MINOR(inode->i_rdev);

        if(minor > 31 || (part = minor%16) >= bd_disk->minors)
        {
                return -EINVAL;
        }

        if( part == 0 )
        {
                /* read/write on whole /dev/sda or /dev/sdb */
                start_sect = 1;
                nr_sects = bd_disk->capacity;
        }
        else
        {
                /* read/write on /dev/sda[1-15] or /dev/sdb[1-3] */
                part = part - 1;

                if(!bd_disk->part[part]) {
                        return -ENXIO;
                }

                if(bd_disk->part[part]->nr_sects == 0) {
                        return -ENXIO;
                }

                start_sect = bd_disk->part[part]->start_sect;
                nr_sects = bd_disk->part[part]->nr_sects;
        }

        memset(&sghdr, 0, sizeof(struct sg_io_hdr));
        memset(tagarray, 0, sizeof tagarray);
        argPtr = (char *) arg;
        nTags = 0;

        /* Process the tags in the array.  Check for out-of-range tag codes and
        * duplicates. */
        do {
                if (copy_from_user(&tagtuple, argPtr, sizeof tagtuple)) {
                        return -EFAULT;
                }
                if (tagtuple.tag < TIVOTAGIO_END || tagtuple.tag >= TIVOTAGIO_NUMTAGS) {
                        return -EINVAL;
                }
                if (tagarray[tagtuple.tag].present++) {
                        return -EINVAL;
                }
                tagarray[tagtuple.tag].tagtuple = tagtuple;
                argPtr += sizeof tagtuple;
                nTags ++;
        } while (tagtuple.tag != TIVOTAGIO_END);

        /* Check for required tags and acceptable values */

        if (!tagarray[TIVOTAGIO_CMD].present ||
        !tagarray[TIVOTAGIO_IOVEC].present ||
        !tagarray[TIVOTAGIO_SECTOR].present ||
        !tagarray[TIVOTAGIO_NRSECTORS].present ||
        (tagarray[TIVOTAGIO_CMD].tagtuple.val != READ &&
        tagarray[TIVOTAGIO_CMD].tagtuple.val != WRITE)) {
                return -EINVAL;
        }

        rw = tagarray[TIVOTAGIO_CMD].tagtuple.val;
        
        /* Offset for the beginning of this partition */
        sector = tagarray[TIVOTAGIO_SECTOR].tagtuple.val;
        nr_sectors = tagarray[TIVOTAGIO_NRSECTORS].tagtuple.val;

        /* Check to see if its off the end of this partition */
        if (sector > start_sect+nr_sects ||
        (sector + nr_sectors) > start_sect+nr_sects) {
                printk(KERN_WARNING "scsi_dosectors: Access beyond end of "
                                "drive %llu %lu %llu %lu\n",sector,start_sect,nr_sectors,nr_sects);
                return -EINVAL; 
        }

        user_len = tagarray[TIVOTAGIO_IOVEC].tagtuple.size/sizeof(struct FsIovec);
        iov_user = (const struct FsIovec *) (unsigned long int)
                        tagarray[TIVOTAGIO_IOVEC].tagtuple.val;
        sector += start_sect;

        /*
         * support non-monotonic (gettimeofday-based) deadlines for backwards
         * compatability, by converting the time diff to a monotonic time
         */
        if (tagarray[TIVOTAGIO_DEADLINE].present &&
                        tagarray[TIVOTAGIO_DEADLINE].tagtuple.val != 0) {
                long long delta_usec;
                delta_usec = tagarray[TIVOTAGIO_DEADLINE].tagtuple.val - now_usec();
                tivo_monotonic_get_current(&deadline);
                tivo_monotonic_add_usecs(&deadline, delta_usec);
        } else {
                /*
                 * yes, this uses 0 to mean no deadline, which leaves a small possibility
                 * every thousand years or so of system run time that a real deadline
                 * will be ignored.  Live with it.
                 */
                tivo_monotonic_store(&deadline,
                                tagarray[TIVOTAGIO_DEADLINE_MONOTONIC].tagtuple.val);
        }

        if (!access_ok(VERIFY_READ, iov_user, user_len)) {
                printk( "Fail check iovec %08x\n", (unsigned int)iov_user );
                return -EFAULT;
        }

        if(nr_sectors <= MAX_NRSECTORS)
        {
                /*No need to split the request, send it directly */
                /* prepare a real request */
                prepare_scsi_request(&sghdr, sector, nr_sectors, iov_user, user_len, rw);
                return scsi_dosectors_io(filp,bd_disk,&sghdr,sector,nr_sectors,&deadline);
        }
        else
        {
                nbytes = split_requests(filp, bd_disk, iov_user, user_len , 
                                sector, &deadline, rw);

                return nbytes;
        }
}

static void prepare_scsi_request(struct sg_io_hdr *sg_hdr, unsigned long long sector,unsigned long long nr_sectors,
                const struct FsIovec *iov_user, int len, int direction)
{
        struct sg_io_hdr *io_hdr = sg_hdr;
        unsigned char rdCmd[BLK_MAX_CDB];
        unsigned char senseBuff[SCSI_SENSE_BUFFERSIZE];

        /*Build read/write command*/
        memset(rdCmd, 0, BLK_MAX_CDB);

        rdCmd[0] = (unsigned char)(direction == READ ? READ_16 : WRITE_16);
        rdCmd[2] = (unsigned char)((sector >> 56) & 0xff);
        rdCmd[3] = (unsigned char)((sector >> 48) & 0xff);
        rdCmd[4] = (unsigned char)((sector >> 40) & 0xff);
        rdCmd[5] = (unsigned char)((sector >> 32) & 0xff);
        rdCmd[6] = (unsigned char)((sector >> 24) & 0xff);
        rdCmd[7] = (unsigned char)((sector >> 16) & 0xff);
        rdCmd[8] = (unsigned char)((sector >> 8) & 0xff);
        rdCmd[9] = (unsigned char)(sector & 0xff);
        rdCmd[10] = (unsigned char)((nr_sectors >> 24) & 0xff);
        rdCmd[11] = (unsigned char)((nr_sectors >> 16) & 0xff);
        rdCmd[12] = (unsigned char)((nr_sectors >> 8) & 0xff);
        rdCmd[13] = (unsigned char)(nr_sectors & 0xff);

        io_hdr->interface_id = 'S';
        io_hdr->cmd_len = 16;
        io_hdr->cmdp = rdCmd;
        if ( direction == READ )
                io_hdr->dxfer_direction = SG_DXFER_FROM_DEV;
        else
                io_hdr->dxfer_direction = SG_DXFER_TO_DEV;
        io_hdr->dxfer_len = nr_sectors; /* in sectors */
        io_hdr->dxferp = iov_user;
        io_hdr->iovec_count = len;
        io_hdr->mx_sb_len = SCSI_SENSE_BUFFERSIZE;
        io_hdr->sbp = senseBuff;
} 

static int split_requests(struct file *filp, struct gendisk *bd_disk, const struct FsIovec *iovec,
                                int iovec_count, unsigned long long start_sect,
                                monotonic_t *deadline, int rw)
{
    const struct FsIovec *curr_iovec;
    unsigned int cb = 0;
    unsigned int byte_count, count, nr_iovec;
    int bytes_transferred, bytes, split_iovec;
    sector_t target_sect, n_sect;
    struct sg_io_hdr sghdr;
    struct sg_iovec iov_table;

    byte_count = count = 0;
    bytes_transferred = bytes = split_iovec = 0;

    target_sect = start_sect;

    while (count < iovec_count)
    {
            curr_iovec = iovec;
            nr_iovec = 0;
            n_sect = 0;
            byte_count = 0;

            for (; count < iovec_count ; count++) {
                    if (__get_user(cb, &iovec->cb)) {
                            printk( "Fail read iovec %08x\n", (unsigned int)iovec );
                            return -EFAULT;
                    }

                    if((byte_count + cb) > MAX_BYTES){

                        if(byte_count > 0) {
                            break;
                        }
                        else if(byte_count == 0 && cb > MAX_BYTES){
                                /* A single iovec of size > MAX_NRSECTORS */

                                int length;
                                unsigned int *base;
                                void *pb;
                                split_iovec = 1;

                                if (__get_user(pb, &iovec->pb)) {
                                        printk( "Fail read iovec %08x\n", (unsigned int)iovec );
                                        return -EFAULT;
                                }

                                base = (unsigned int *) pb;
                                length = cb;

                                while(length > 0){
                                        iov_table.iov_base = base;

                                        if(length >= MAX_BYTES) {
                                                iov_table.iov_len = MAX_BYTES; 
                                        }
                                        else {
                                                iov_table.iov_len = length;
                                        }

                                        base += MAX_BYTES/4;
                                        length -= MAX_BYTES;

                                        if(iov_table.iov_len % 512){
                                                printk("scsi_dosector: iovec bytes not aligned on sector boundary\n");
                                                if(!(err_count--))
                                                BUG();
                                        }

                                        n_sect = iov_table.iov_len / SECTOR_BYTES;

                                        memset(&sghdr, 0, sizeof(struct sg_io_hdr));

                                        prepare_scsi_request(&sghdr, target_sect, n_sect, 
                                                        (struct FsIovec *)&iov_table, 1, rw);

                                        bytes = scsi_dosectors_io(filp, bd_disk, &sghdr,
                                                        target_sect, n_sect, deadline);

                                        if (bytes < 0) {
                                                goto out;
                                        }
                                        else {
                                                bytes_transferred += bytes;
                                        }

                                        target_sect += n_sect;
                                }
                                        iovec++;
                                        continue;
                        }
                    }
                    byte_count += cb;
                    iovec++;
                    nr_iovec++; 
            }

            if(split_iovec){
                    split_iovec = 0;
                    continue;
            }

            if((byte_count % SECTOR_BYTES)){
                    printk("scsi_dosector: iovec bytes not aligned on sector boundary\n");
                    if(!(err_count--))
                    BUG();
            }

            n_sect = byte_count / SECTOR_BYTES;

            memset(&sghdr, 0, sizeof(struct sg_io_hdr));

            prepare_scsi_request(&sghdr, target_sect, n_sect, 
                            curr_iovec, nr_iovec, rw);

            bytes = scsi_dosectors_io(filp, bd_disk, &sghdr,
                            target_sect, n_sect, deadline);

            if (bytes < 0) {
                    break;
            }
            else {
                    bytes_transferred += bytes;
            }

            target_sect += n_sect;
    }

out:
    return bytes_transferred;
}

